/*
 * All of the documentation and software included in the
 * Alchemy Software is copyrighted by Stanley Kok, Parag
 * Singla, Matthew Richardson, Pedro Domingos, Marc
 * Sumner and Hoifung Poon.
 * 
 * Copyright [2004-06] Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon. All rights reserved.
 * 
 * Contact: Pedro Domingos, University of Washington
 * (pedrod@cs.washington.edu).
 * 
 * Redistribution and use in source and binary forms, with
 * or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above
 * copyright notice, this list of conditions and the
 * following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the
 * above copyright notice, this list of conditions and the
 * following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use
 * of this software must display the following
 * acknowledgment: "This product includes software
 * developed by Stanley Kok, Parag Singla, Matthew
 * Richardson, Pedro Domingos, Marc Sumner and Hoifung
 * Poon in the Department of Computer Science and
 * Engineering at the University of Washington".
 * 
 * 4. Your publications acknowledge the use or
 * contribution made by the Software to your research
 * using the following citation(s): 
 * Stanley Kok, Parag Singla, Matthew Richardson and
 * Pedro Domingos (2005). "The Alchemy System for
 * Statistical Relational AI", Technical Report,
 * Department of Computer Science and Engineering,
 * University of Washington, Seattle, WA.
 * http://www.cs.washington.edu/ai/alchemy.
 * 
 * 5. Neither the name of the University of Washington nor
 * the names of its contributors may be used to endorse or
 * promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY
 * OF WASHINGTON OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
#ifndef VOTED_PERCEPTRON_H_OCT_30_2005
#define VOTED_PERCEPTRON_H_OCT_30_2005

#include "infer.h"
#include "clause.h"
#include "timer.h"
#include "indextranslator.h"
#include "maxwalksat.h"

const bool vpdebug = false;
const double EPSILON=.00001;

/**
 * VotedPerceptron algorithm (see "Discriminative Training of Markov Logic
 * Networks", Singla and Domingos, 2005).
 */
class VotedPerceptron 
{
 public:

  /**
   * Constructor.
   */
  VotedPerceptron(const Array<MLN*>& mlns, const Array<Domain*>& domains,
                  const string& nonEvidPredsStr, const int& maxSteps,
                  const int& tries, const long double& targetWt,
                  const bool& hard, IndexTranslator* const & idxTrans,
                  const int& memLimit, const bool& lazyInference,
                  const bool& rescaleGradient, const bool& withEM)
    : domainCnt_(domains.size()), idxTrans_(idxTrans)
  { 
    cout << endl << "Constructing voted perceptron..." << endl << endl;
    //cout << "domain count = " << domainCnt_ << endl;

    assert(mlns.size() == domains.size());
    domains_.append(domains);
    mlns_.append(mlns);

    logOddsPerDomain_.growToSize(domainCnt_);
    clauseCntPerDomain_.growToSize(domainCnt_);
    
    for (int i = 0; i < domainCnt_; i++)
    {
      clauseCntPerDomain_[i] = mlns[i]->getNumClauses();
      logOddsPerDomain_[i].growToSize(clauseCntPerDomain_[i], 0);

      //cout<< "number of clauses in domain " << i << " = "
      //    << clauseCntPerDomain_[i] <<endl;
    }

	memLimit_ = memLimit;
	lazyInference_ = lazyInference;
	lazyReset_ = true;
    rescaleGradient_ = rescaleGradient;
	withEM_ = withEM;

    if (!lazyInference_)
    {
        // Remove evidence atoms structure from DBs
      for (int i = 0; i < domainCnt_; i++)
      {
        domains_[i]->getDB()->setLazyFlag(false);
      }
    }
    
    totalTrueCnts_.growToSize(domainCnt_);
    defaultTrueCnts_.growToSize(domainCnt_);
    relevantClausesPerDomain_.growToSize(domainCnt_);
    states_.growToSize(domainCnt_);
    inferences_.growToSize(domainCnt_);
    //relevantClausesFormulas_ is set in findRelevantClausesFormulas()

      // Unknown evidence
    uePredsPerDomain_.growToSize(domainCnt_);

      // Known evidence
    kePredsPerDomain_.growToSize(domainCnt_);
    kePredValuesPerDomain_.growToSize(domainCnt_);    

      // Unknown non-evidence
    unePredsPerDomain_.growToSize(domainCnt_);
    uneEMPredsPerDomain_.growToSize(domainCnt_);

      // Known non-evidence
    knePredsPerDomain_.growToSize(domainCnt_);
    knePredValuesPerDomain_.growToSize(domainCnt_);

    if (withEM_)
    {
      mvStates_.growToSize(domainCnt_);
      mvInferences_.growToSize(domainCnt_);
    }

    StringHashArray nonEvidPredNames;
      //defined in infer.h
    extractPredNames(nonEvidPredsStr, NULL, nonEvidPredNames);
    findRelevantClauses(nonEvidPredNames);
    findRelevantClausesFormulas();

      // TODO: Allow user-set parameters
      // Intialize the maxwalksat parameters
    wsparams_ = NULL;
    wsparams_ = new MaxWalksatParams;
    wsparams_->maxSteps = maxSteps;
    wsparams_->maxTries = tries;
    wsparams_->targetCost = targetWt;
    wsparams_->hard = hard;
    wsparams_->numSolutions = 1;
      // TABU heuristic
    wsparams_->heuristic = 2;
    wsparams_->tabuLength = 10;

      // Initialize the clause wts for lazy version
    if (lazyInference_)
    {
      findCountsInitializeWtsAndSetNonEvidPredsToUnknownInDB(nonEvidPredNames);
    
      for (int i = 0; i < mlns_.size(); i++)
      {
        MLN* mln = mlns_[i];
        Array<double>& logOdds = logOddsPerDomain_[i];
        assert(mln->getNumClauses() == logOdds.size());
        for (int j = 0; j < mln->getNumClauses(); j++)
          ((Clause*) mln->getClause(j))->setWt(logOdds[j]);
      }
    }

      // Build the state for inference in each domain
    Array<int> allPredGndingsAreNonEvid;
    bool markHardGndClauses = false;
    bool trackParentClauseWts = true;

      // Hold known and unknown non-evidence preds
    GroundPredicateHashArray knownNEGndPreds;
    GroundPredicateHashArray unknownNEGndPreds;

    Array<Predicate *> nonEvidPreds;
    Array<TruthValue> nonEvidPredValues;

    Array<Predicate*> gpreds;
    Array<Predicate*> ppreds;
    Array<TruthValue> gpredValues;
    Array<TruthValue> tmpValues;
    
      // Unknown evidence preds
    Array<Predicate*> uepreds;    
    Array<TruthValue> uepredValues;

      // unknownGndPreds will hold all unknown preds (evidence and non-evidence)
    GroundPredicateHashArray unknownGndPreds;
    GroundPredicateHashArray nonEvidenceGndPreds;

    for (int i = 0; i < domainCnt_; i++)
    {
      Domain* domain = domains_[i];
      MLN* mln = mlns_[i];

        // Unknown ev. preds
      GroundPredicateHashArray& uePreds = uePredsPerDomain_[i];

        // Known ev. preds
      GroundPredicateHashArray& kePreds = kePredsPerDomain_[i];
      Array<TruthValue>& kePredValues = kePredValuesPerDomain_[i];

        // Unknown non-ev. preds
      GroundPredicateHashArray& unePreds = unePredsPerDomain_[i];
        // Unknown non-ev. preds for missing value state
      GroundPredicateHashArray& uneEMPreds = uneEMPredsPerDomain_[i];

        // Known non-ev. preds
      GroundPredicateHashArray& knePreds = knePredsPerDomain_[i];
      Array<TruthValue>& knePredValues = knePredValuesPerDomain_[i];

      VariableState*& state = states_[i];
      Inference*& inference = inferences_[i];
      
        // Need to set some dummy weight
      for (int j = 0; j < mln->getNumClauses(); j++)
        ((Clause*) mln->getClause(j))->setWt(1);

      if (!lazyInference_)
      {
        //knownPreds.clear();
        knownNEGndPreds.clear();
        unknownNEGndPreds.clear();

        allPredGndingsAreNonEvid.growToSize(domain->getNumPredicates(), false);
          //defined in infer.h
        createComLineQueryPreds(nonEvidPredsStr, domain, domain->getDB(), 
                                &unePreds, &knePreds, 
                                &allPredGndingsAreNonEvid);
        //cout << "created the com line query .." << endl;

          // Pred values not set to unknown in DB: unePreds contains
          // unknown, knePreds contains known non-evidence

          // Copy unePreds for missing value state
        for (int predno = 0; predno < unePreds.size(); predno++) 
        {
          Predicate* p = unePreds[predno]->createEquivalentPredicate(domain);
          uneEMPreds.append(new GroundPredicate(p));
          delete p;
        }
        
          // Find all evidence predicates with unknown value for VPEM
        int numPreds = domain->getNumPredicates();
        for (int j = 0; j < numPreds; j++)
        {
            // Only look at evidence (but not '=' predicate)
          if (nonEvidPredNames.contains(domain->getPredicateName(j)) ||
              domain->getPredicateTemplate(j)->isEqualPredicateTemplate())
            continue;
          
          ppreds.clear();
          Predicate::createAllGroundings(j, domain, ppreds);
          uepreds.append(ppreds);
        }

        uePreds.deleteItemsAndClear();
        kePreds.deleteItemsAndClear();
        kePredValues.clear();
        for (int j = 0; j < uepreds.size(); j++)
        {
          TruthValue tv = domain->getDB()->getValue(uepreds[j]);
          Predicate* pred = new Predicate(*(uepreds[j]));
          GroundPredicate* gpred = new GroundPredicate(pred);
          if (tv == UNKNOWN)
          {
            uePreds.append(gpred);
          }
          else
          {
            kePreds.append(gpred);
            kePredValues.append(tv);
          }
        }

        unknownGndPreds.append(uneEMPreds);
        unknownGndPreds.append(uePreds);
          // unknownGndPreds now holds all unknown preds (evidence and
          // non-evidence)

        if (withEM_)
        {          
          cout << "creating missing value state ..." << endl;
            // Build the state to fill in missing values based on this
          mvStates_[i] = new VariableState(&unknownGndPreds, NULL, 
                                           markHardGndClauses,
                                           trackParentClauseWts, mln, domain,
                                           lazyInference_);
          int seed = -1;
          mvInferences_[i] = new MaxWalkSat(mvStates_[i], seed, wsparams_);
          cout << "created missing value state" << endl;
        }
          // Need to set unknownPreds to UNKNOWN when doing inference

          // Create state for inferred counts using unknown and known (set to
          // unknown in the DB) non-evidence preds
        cout << endl << "constructing state for domain " << i << "..."<<endl;

          // Set unknown evidence to FALSE for building state
        for (int predno = 0; predno < uePreds.size(); predno++) 
        {
          domain->getDB()->setValue(uePreds[predno], FALSE);
        }

          // Set known NE to unknown for building state
        knePredValues.growToSize(knePreds.size(), FALSE);
        for (int predno = 0; predno < knePreds.size(); predno++) 
        {
          knePredValues[predno] =
            domain->getDB()->setValue(knePreds[predno], UNKNOWN);
        }

        nonEvidenceGndPreds.append(knePreds);
        nonEvidenceGndPreds.append(unePreds);
      }
        
        // Create inference algorithm and state based on queries and
        // mln / domain
      state = new VariableState(&nonEvidenceGndPreds, &allPredGndingsAreNonEvid,
                                markHardGndClauses, trackParentClauseWts,
                                mln, domain, lazyInference_);
      int seed = -1;
      inference = new MaxWalkSat(state, seed, wsparams_);

      if (!lazyInference_)
      {
          // Change known NE to original values
        domains_[i]->getDB()->setValuesToGivenValues(&knePreds,
                                                     &knePredValues);
          // Set unknown NE to false for weight initialization
          // This seems to give poor results. We need to leave these
          // as unknown and do the counts accordingly
        for (int predno = 0; predno < unePreds.size(); predno++) 
        {
          domain->getDB()->setValue(unePreds[predno], FALSE);
        }
        
          // Clean up        
        for (int predno = 0; predno < uepreds.size(); predno++) 
          delete uepreds[predno];
      }
        
        // Initialize the inference / state
      //inference->init();
      //state->initRandom();
    }
    cout << endl << "done constructing variable states" << endl << endl;

      // Initialize the clause wts for eager version
	if (!lazyInference_)
	{      
	  initializeWts();
    }

  }


  ~VotedPerceptron() 
  {
    for (int i = 0; i < inferences_.size(); i++) delete inferences_[i];
    for (int i = 0; i < states_.size(); i++) delete states_[i];

    if (withEM_)
    {
      for (int i = 0; i < mvInferences_.size(); i++)
      {
        delete mvInferences_[i];
        delete mvStates_[i];
      }
    }
    
    for (int i = 0; i < trainTrueCnts_.size(); i++)
      delete[] trainTrueCnts_[i];

    delete wsparams_;
  }


    // set the prior means and std devs.
  void setMeansStdDevs(const int& arrSize, const double* const & priorMeans, 
                       const double* const & priorStdDevs) 
  {
    if (arrSize < 0) 
    {
      usePrior_ = false;
      priorMeans_ = NULL;
      priorStdDevs_ = NULL;
    } 
    else 
    {
      //cout << "arr size = " << arrSize<<", clause count = "<<clauseCnt_<<endl;
      usePrior_ = true;
      priorMeans_ = priorMeans;
      priorStdDevs_ = priorStdDevs;

      //cout << "\t\t Mean \t\t Std Deviation" << endl;
      //for (int i = 0; i < arrSize; i++) 
      //  cout << i << "\t\t" << priorMeans_[i]<<"\t\t"<<priorStdDevs_[i]<<endl;
    }
  }


    // learn the weights
  void learnWeights(double* const & weights, const int& numWeights,
                    const int& maxIter, const double& learningRate,
                    const double& momentum, bool initWithLogOdds) 
  {
    //cout << "Learning weights discriminatively... " << endl;
    memset(weights, 0, numWeights*sizeof(double));

    double* averageWeights = new double[numWeights];
    double* gradient = new double[numWeights];
    double* lastchange = new double[numWeights];

      // Set the initial weight to the average log odds across domains/databases
    if (initWithLogOdds)
    {
        // If there is one db or the clauses for multiple databases line up
      if (idxTrans_ == NULL)
      {
        for (int i = 0; i < domains_.size(); i++)
        {
          Array<double>& logOdds = logOddsPerDomain_[i];
          assert(numWeights == logOdds.size());
          for (int j = 0; j < logOdds.size(); j++) weights[j] += logOdds[j];
        }
      }
      else
      { //the clauses for multiple databases do not line up
        const Array<Array<Array<IdxDiv>*> >* cIdxToCFIdxsPerDomain 
          = idxTrans_->getClauseIdxToClauseFormulaIdxsPerDomain();

        Array<int> numLogOdds; 
        Array<double> wtsForDomain;
        numLogOdds.growToSize(numWeights);
        wtsForDomain.growToSize(numWeights);
      
        for (int i = 0; i < domains_.size(); i++)
        {
          memset((int*)numLogOdds.getItems(), 0, numLogOdds.size()*sizeof(int));
          memset((double*)wtsForDomain.getItems(), 0,
                 wtsForDomain.size()*sizeof(double));

          Array<double>& logOdds = logOddsPerDomain_[i];
        
            // Map the each log odds of a clause to the weight of a
            // clause/formula
          for (int j = 0; j < logOdds.size(); j++)
          {
            Array<IdxDiv>* idxDivs =(*cIdxToCFIdxsPerDomain)[i][j];          
            for (int k = 0; k < idxDivs->size(); k++)
            {
              wtsForDomain[ (*idxDivs)[k].idx ] += logOdds[j];
              numLogOdds[ (*idxDivs)[k].idx ]++;
            }
          }

          for (int j = 0; j < numWeights; j++)
            if (numLogOdds[j] > 0) weights[j] += wtsForDomain[j]/numLogOdds[j];  
        }
      }
    }

      // Initialize weights, averageWeights, lastchange
    for (int i = 0; i < numWeights; i++) 
    {      
      weights[i] /= domains_.size();
      averageWeights[i] = weights[i];
      lastchange[i] = 0.0;
    }

    for (int iter = 1; iter <= maxIter; iter++) 
    {
      cout << endl << "Iteration " << iter << " : " << endl << endl;
      cout << "Getting the gradient.. " << endl;
      getGradient(weights, gradient, numWeights);
      cout << endl; 

        // Add gradient to weights
      for (int w = 0; w < numWeights; w++) 
      {
        double wchange = gradient[w] * learningRate + lastchange[w] * momentum;
        cout << "clause/formula " << w << ": wtChange = " << wchange;
        cout << "  oldWt = " << weights[w];
        double oldWt = weights[w];
        weights[w] += wchange;
        lastchange[w] = wchange;
        cout << "  newWt = " << weights[w];
          // If weight of one clause goes from neg. to pos. or vice versa,
          // then reset lazysat
        if ((oldWt < 0 && weights[w] > 0) || (oldWt > 0 && weights[w] < 0))
          lazyReset_ = true;
        averageWeights[w] = (iter * averageWeights[w] + weights[w])/(iter + 1);
        cout << "  averageWt = " << averageWeights[w] << endl;
      }
      // done with an iteration
    }
    
    cout << endl << "Learned Weights : " << endl;
    for (int w = 0; w < numWeights; w++) 
    {
      weights[w] = averageWeights[w];
      cout << w << ":" << weights[w] << endl;
    }

    delete [] averageWeights;
    delete [] gradient;
    delete [] lastchange;
    
    resetDBs();
  }
 
 
 private:
 
  /**
   * Resets the values of non-evidence predicates as they were before learning.
   */
  void resetDBs() 
  {
    for(int i = 0; i < domainCnt_; i++) 
    {
      Database* db = domains_[i]->getDB();
        // Change known NE to original values
      db->setValuesToGivenValues(&(knePredsPerDomain_[i]),
                                 &(knePredValuesPerDomain_[i]));
        // Set unknown NE back to UKNOWN
      for (int predno = 0; predno < unePredsPerDomain_[i].size(); predno++) 
      {
        db->setValue(unePredsPerDomain_[i][predno], UNKNOWN);
      }
    }
  }

  /**
   * Assign true to the elements in the relevantClauses_ bool array 
   * corresponding to indices of clauses which would be relevant for list of 
   * non-evidence predicates.
   */
  void findRelevantClauses(const StringHashArray& nonEvidPredNames) 
  {
    for (int d = 0; d < domainCnt_; d++)
    {
      int clauseCnt = clauseCntPerDomain_[d];
      Array<bool>& relevantClauses = relevantClausesPerDomain_[d];
      relevantClauses.growToSize(clauseCnt);
      memset((bool*)relevantClauses.getItems(), false, 
             relevantClauses.size()*sizeof(bool));
      Domain* domain = domains_[d];
      MLN* mln = mlns_[d];
    
      const Array<IndexClause*>* indclauses;
      const Clause* clause;
      int predid, clauseid;
      for (int i = 0; i < nonEvidPredNames.size(); i++)
      {
        predid = domain->getPredicateId(nonEvidPredNames[i].c_str());
        //cout << "finding the relevant clauses for predid = " << predid 
        //     << " in domain " << d << endl;
        indclauses = mln->getClausesContainingPred(predid);
        if (indclauses) 
        {
          for (int j = 0; j < indclauses->size(); j++) 
          {
            clause = (*indclauses)[j]->clause;			
            clauseid = mln->findClauseIdx(clause);
            relevantClauses[clauseid] = true;
            //cout << clauseid << " ";
          }
          //cout<<endl;
        }
      }    
    }
  }

  
  void findRelevantClausesFormulas()
  {
    if (idxTrans_ == NULL)
    {
      Array<bool>& relevantClauses = relevantClausesPerDomain_[0];
      relevantClausesFormulas_.growToSize(relevantClauses.size());
      for (int i = 0; i < relevantClauses.size(); i++)
        relevantClausesFormulas_[i] = relevantClauses[i];
    }
    else
    {
      idxTrans_->setRelevantClausesFormulas(relevantClausesFormulas_,
                                            relevantClausesPerDomain_[0]);
      cout << "Relevant clauses/formulas:" << endl;
      idxTrans_->printRelevantClausesFormulas(cout, relevantClausesFormulas_);
      cout << endl;
    }
  }


  /**
   * Calculate true/false/unknown counts for all clauses for the given domain.
   * 
   * @param trueCnt Number of true groundings for each clause is stored here.
   * @param falseCnt Number of false groundings for each clause is stored here.
   * @param domainIdx Index of domain where the groundings are counted.
   * @param hasUnknownPreds If true, the domain has predicates with unknown
   * truth values. Otherwise it contains only predicates with known values.
   */
  void calculateCounts(Array<double>& trueCnt, Array<double>& falseCnt,
                       const int& domainIdx, const bool& hasUnknownPreds) 
  {
    Clause* clause;
    double tmpUnknownCnt;
    int clauseCnt = clauseCntPerDomain_[domainIdx];
    Array<bool>& relevantClauses = relevantClausesPerDomain_[domainIdx];
    MLN* mln = mlns_[domainIdx];
    Domain* domain = domains_[domainIdx];

    for (int clauseno = 0; clauseno < clauseCnt; clauseno++) 
    {
      if (!relevantClauses[clauseno]) 
      {
        continue;
        //cout << "\n\nthis is an irrelevant clause.." << endl;
      }
      //cout << clauseno << ": ";
      clause = (Clause*) mln->getClause(clauseno);
      clause->getNumTrueFalseUnknownGroundings(domain, domain->getDB(), 
                                               hasUnknownPreds,
                                               trueCnt[clauseno],
                                               falseCnt[clauseno],
                                               tmpUnknownCnt);
      //cout << "true = " << trueCnt[clauseno] << " ";
      //cout << "false = " << falseCnt[clauseno] << " ";
      //cout << "unknown = " << tmpUnknownCnt << endl;
      assert(hasUnknownPreds || (tmpUnknownCnt==0));
    }
  }


  void initializeWts()
  {
    cout << "Initializing weights ..." << endl;
    Array<double *> trainFalseCnts;
    trainTrueCnts_.growToSize(domainCnt_);
    trainFalseCnts.growToSize(domainCnt_);
  
    for (int i = 0; i < domainCnt_; i++)
    {
      int clauseCnt = clauseCntPerDomain_[i];
      GroundPredicateHashArray& unePreds = unePredsPerDomain_[i];
      GroundPredicateHashArray& knePreds = knePredsPerDomain_[i];

      trainTrueCnts_[i] = new double[clauseCnt];
      trainFalseCnts[i] = new double[clauseCnt];
      
      int totalPreds = knePreds.size() + unePreds.size();
        // Used to store gnd preds to be ignored in the count because they are
        // UNKNOWN
      Array<bool>* unknownPred = new Array<bool>;
      unknownPred->growToSize(totalPreds, false);
      for (int predno = 0; predno < totalPreds; predno++) 
      {
        GroundPredicate* p;
        if (predno < knePreds.size())
          p = knePreds[predno];
        else
          p = unePreds[predno - knePreds.size()];
        TruthValue tv = domains_[i]->getDB()->getValue(p);
        //assert(tv != UNKNOWN);
        if (tv == TRUE)
        {
          states_[i]->setValueOfAtom(predno + 1, true);
          p->setTruthValue(true);
        }
        else
        {
          states_[i]->setValueOfAtom(predno + 1, false);
          p->setTruthValue(false);
            // Can have unknown truth values when using EM. We want to ignore
            // these when performing the counts
          if (tv == UNKNOWN)
          {
            (*unknownPred)[predno] = true;
          }
        }
      }

      states_[i]->initMakeBreakCostWatch(0);
      //cout<<"getting true cnts => "<<endl;
      states_[i]->getNumClauseGndingsWithUnknown(trainTrueCnts_[i], clauseCnt,
                                                 true, unknownPred);
      //cout<<endl;
      //cout<<"getting false cnts => "<<endl;
      states_[i]->getNumClauseGndingsWithUnknown(trainFalseCnts[i], clauseCnt,
                                                 false, unknownPred);
      delete unknownPred;
      if (vpdebug)
      {
        for (int clauseno = 0; clauseno < clauseCnt; clauseno++)
        {
          cout << clauseno << " : tc = " << trainTrueCnts_[i][clauseno]
               << " ** fc = " << trainFalseCnts[i][clauseno] << endl;
        }
      }
    }

    double tc,fc;
    cout << "List of CNF Clauses : " << endl;
    for (int clauseno = 0; clauseno < clauseCntPerDomain_[0]; clauseno++)
    {
      if (!relevantClausesPerDomain_[0][clauseno])
      {
        for (int i = 0; i < domainCnt_; i++)
        {
          Array<double>& logOdds = logOddsPerDomain_[i];
          logOdds[clauseno] = 0.0;
        }
        continue;
      }
      //cout << endl << endl;
      cout << clauseno << ":";
      const Clause* clause = mlns_[0]->getClause(clauseno);
      //cout << (*fncArr)[clauseno]->formula <<endl;
      clause->print(cout, domains_[0]);
      cout << endl;
      
      tc = 0.0; fc = 0.0;
      for (int i = 0; i < domainCnt_;i++)
      {
        tc += trainTrueCnts_[i][clauseno];
        fc += trainFalseCnts[i][clauseno];
      }
	
      //cout << "true count  = " << tc << endl;
      //cout << "false count = " << fc << endl;
	
      double weight = 0.0;
      double totalCnt = tc + fc;
		
      if (totalCnt == 0) 
      {
        //cout << "NOTE: Total count is 0 for clause " << clauseno << endl;
        weight = EPSILON;
      } 
      else 
      {
        double prob =  tc / (tc+fc);
        if (prob == 0) prob = 0.00001;
        if (prob == 1) prob = 0.99999;
        weight = log(prob/(1-prob));
          //if weight exactly equals 0, make it small non zero, so that clause  
          //is not ignored during the construction of the MRF
        //if(weight == 0) weight = 0.0001;
          //commented above - make sure all weights are positive in the
          //beginning
        //if(weight < EPSILON) weight = EPSILON;
        if (abs(weight) < EPSILON) weight = EPSILON;
          //cout << "Prob " << prob << " becomes weight of " << weight << endl;
      }
      for (int i = 0; i < domainCnt_; i++) 
      {
      	Array<double>& logOdds = logOddsPerDomain_[i];
        logOdds[clauseno] = weight;
      }
    }
    cout << endl;
    
    for (int i = 0; i < trainFalseCnts.size(); i++)
      delete[] trainFalseCnts[i];
  }

  /**
   * Finds the training counts and intialize the weights.
   */
  void findCountsInitializeWtsAndSetNonEvidPredsToUnknownInDB(
                                       const StringHashArray& nonEvidPredNames)
  {
    bool hasUnknownPreds;
    Array<Array<double> > totalFalseCnts; 
    Array<Array<double> > defaultFalseCnts;
    totalFalseCnts.growToSize(domainCnt_);
    defaultFalseCnts.growToSize(domainCnt_);
    
    Array<Predicate*> gpreds;
    Array<Predicate*> ppreds;
    Array<TruthValue> gpredValues;
    Array<TruthValue> tmpValues;

    for (int i = 0; i < domainCnt_; i++) 
    {
      Domain* domain = domains_[i];
      int clauseCnt = clauseCntPerDomain_[i];

      //cout << endl << "Getting the counts for the domain " << i << endl;
      gpreds.clear();
      gpredValues.clear();
      tmpValues.clear();
      for (int predno = 0; predno < nonEvidPredNames.size(); predno++) 
      {
        ppreds.clear();
        int predid = domain->getPredicateId(nonEvidPredNames[predno].c_str());
        Predicate::createAllGroundings(predid, domain, ppreds);
        //cout<<"size of gnd for pred " << predid << " = "<<ppreds.size()<<endl;
        gpreds.append(ppreds);
      }
      
      domain->getDB()->alterTruthValue(&gpreds, UNKNOWN, FALSE, &gpredValues);
	  
      //cout <<"size of unknown set for domain "<<i<<" = "<<gpreds.size()<<endl;
      //cout << "size of the values " << i << " = " << gpredValues.size()<<endl;
	
      hasUnknownPreds = false;
      
      Array<double>& trueCnt = totalTrueCnts_[i];
      Array<double>& falseCnt = totalFalseCnts[i];
      trueCnt.growToSize(clauseCnt);
      falseCnt.growToSize(clauseCnt);
      calculateCounts(trueCnt, falseCnt, i, hasUnknownPreds);

      //cout << "got the total counts..\n\n\n" << endl;
	  
      hasUnknownPreds = true;

      domain->getDB()->setValuesToUnknown(&gpreds, &tmpValues);

      Array<double>& dTrueCnt = defaultTrueCnts_[i];
      Array<double>& dFalseCnt = defaultFalseCnts[i];
      dTrueCnt.growToSize(clauseCnt);
      dFalseCnt.growToSize(clauseCnt);
      calculateCounts(dTrueCnt, dFalseCnt, i, hasUnknownPreds);

      //cout<< "size of unknown set for domain " << i << " = " 
      //    << gpreds.size() << endl;
      //cout<< "size of corresponding value set " << i <<" = "
      //    << gpredValues.size() << endl;

      //commented out: no need to revert the grounded non-evidence predicates
      //               to their initial values because we want to set ALL of
      //               them to UNKNOWN
      //assert(gpreds.size() == gpredValues.size());
      //domain->getDB()->setValuesToGivenValues(&gpreds, &gpredValues);
	  
      //cout << "the ground predicates are :" << endl;
      for (int predno = 0; predno < gpreds.size(); predno++) 
        delete gpreds[predno];
    }
    //cout << endl << endl;
    //cout << "got the default counts..." << endl;     
    for (int clauseno = 0; clauseno < clauseCntPerDomain_[0]; clauseno++) 
    {
      double tc = 0;
      double fc = 0;
      for (int i = 0; i < domainCnt_; i++) 
      {
      	Array<bool>& relevantClauses = relevantClausesPerDomain_[i];
      	Array<double>& logOdds = logOddsPerDomain_[i];
      
        if (!relevantClauses[clauseno]) { logOdds[clauseno] = 0; continue; }
        tc += totalTrueCnts_[i][clauseno] - defaultTrueCnts_[i][clauseno];
        fc += totalFalseCnts[i][clauseno] - defaultFalseCnts[i][clauseno];

        if (vpdebug)
          cout << clauseno << " : tc = " << tc << " ** fc = "<< fc <<endl;      
      }
      
      double weight = 0.0;

      if ((tc + fc) == 0) 
      {
        //cout << "NOTE: Total count is 0 for clause " << clauseno << endl;
      } 
      else 
      {
        double prob =  tc / (tc+fc);
        if (prob == 0) prob = 0.00001;
        if (prob == 1) prob = 0.99999;
        weight = log(prob / (1-prob));
            //if weight exactly equals 0, make it small non zero, so that clause
            //is not ignored during the construction of the MRF
        //if (weight == 0) weight = 0.0001;
        if (abs(weight) < EPSILON) weight = EPSILON;
          //cout << "Prob " << prob << " becomes weight of " << weight << endl;
      }
      
      	// Set logOdds in all domains to the weight calculated
      for(int i = 0; i < domainCnt_; i++) 
      { 
      	Array<double>& logOdds = logOddsPerDomain_[i];
        logOdds[clauseno] = weight;
      }
    }
  }
 
  
  /**
   * Runs inference using the current set of parameters.
   */
  void infer() 
  {
    for (int i = 0; i < domainCnt_; i++) 
    {
      states_[i]->setGndClausesWtsToSumOfParentWts();
      inferences_[i]->init();
      inferences_[i]->infer();
      states_[i]->saveLowStateToGndPreds();
    }
  }

  /**
   * Infers values for predicates with unknown truth values.
   */
  void fillInMissingValues()
  {
    assert(withEM_);
    cout << "Filling in missing data ..." << endl;
      // Get values of initial unknown preds by producing MAP state of
      // unknown preds given known evidence and non-evidence preds (VPEM)
    Array<Array<TruthValue> > ueValues;
    ueValues.growToSize(domains_.size());
    for (int i = 0; i < domains_.size(); i++)
    {
      Domain* domain = domains_[i];
      GroundPredicateHashArray& uePreds = uePredsPerDomain_[i];
      GroundPredicateHashArray& knePreds = knePredsPerDomain_[i];
      Array<TruthValue> knePredValues = knePredValuesPerDomain_[i];

        // Set unknown evidence values to unknown
      domain->getDB()->setValuesToUnknown(&uePreds, &ueValues[i]);
        // Mark known non-evidence preds as evidence
      domain->getDB()->setValuesToGivenValues(&knePreds, &knePredValues);

        // Infer missing values
      mvStates_[i]->setGndClausesWtsToSumOfParentWts();
      mvInferences_[i]->init();
      mvInferences_[i]->infer();
      mvStates_[i]->saveLowStateToGndPreds();

      if (vpdebug)
      {
        cout << "Inferred following values: " << endl;
        mvInferences_[i]->printProbabilities(cout);
      }
            
        // Compute counts
      if (lazyInference_)
      {
        Array<double>& trueCnt = totalTrueCnts_[i];
        Array<double> falseCnt;
        bool hasUnknownPreds = false;
        falseCnt.growToSize(trueCnt.size());
        calculateCounts(trueCnt, falseCnt, i, hasUnknownPreds);
      }
      else
      {
        int clauseCnt = clauseCntPerDomain_[i];
        mvStates_[i]->initMakeBreakCostWatch(0);
        //cout<<"getting true cnts => "<<endl;
        mvStates_[i]->getNumClauseGndings(trainTrueCnts_[i], clauseCnt, true);
      }

        // Set evidence values back
      assert(uePreds.size() == ueValues[i].size());
      domain->getDB()->setValuesToGivenValues(&uePreds, &ueValues[i]);
        // Set non-evidence values to unknown
      Array<TruthValue> tmpValues;
      tmpValues.growToSize(knePreds.size());
      domain->getDB()->setValuesToUnknown(&knePreds, &tmpValues);
    }
    cout << "Done filling in missing data" << endl;    
  }

  void getGradientForDomain(double* const & gradient, const int& domainIdx)
  {
    Array<bool>& relevantClauses = relevantClausesPerDomain_[domainIdx];
    int clauseCnt = clauseCntPerDomain_[domainIdx];
    double *trainCnts, *inferredCnts;
    trainCnts = NULL;
    inferredCnts = NULL;
    double* clauseTrainCnts = new double[clauseCnt]; 
    double* clauseInferredCnts = new double[clauseCnt];
    double trainCnt, inferredCnt;
    Array<double>& totalTrueCnts = totalTrueCnts_[domainIdx];
    Array<double>& defaultTrueCnts = defaultTrueCnts_[domainIdx];    
    MLN* mln = mlns_[domainIdx];
    Domain* domain = domains_[domainIdx];

    memset(clauseTrainCnts, 0, clauseCnt*sizeof(double));
    memset(clauseInferredCnts, 0, clauseCnt*sizeof(double));

    if (!lazyInference_)
    {
      if (!inferredCnts) inferredCnts = new double[clauseCnt];
      states_[domainIdx]->getNumClauseGndings(inferredCnts, clauseCnt, true);
      trainCnts = trainTrueCnts_[domainIdx];
    }
      //loop over all the training examples
    //cout << "\t\ttrain count\t\t\t\tinferred count" << endl << endl;
    for (int clauseno = 0; clauseno < clauseCnt; clauseno++) 
    {
      if (!relevantClauses[clauseno]) continue;
      
      if (lazyInference_)
      {
      	Clause* clause = (Clause*) mln->getClause(clauseno);

      	trainCnt = totalTrueCnts[clauseno];
      	inferredCnt =
          clause->getNumTrueGroundings(domain, domain->getDB(), false);
      	trainCnt -= defaultTrueCnts[clauseno];
      	inferredCnt -= defaultTrueCnts[clauseno];
      
      	clauseTrainCnts[clauseno] += trainCnt;
      	clauseInferredCnts[clauseno] += inferredCnt;
      }
      else
      {
      	clauseTrainCnts[clauseno] += trainCnts[clauseno];
      	clauseInferredCnts[clauseno] += inferredCnts[clauseno];
      }
      //cout << clauseno << ":\t\t" <<trainCnt<<"\t\t\t\t"<<inferredCnt<<endl;
    }

    if (vpdebug)
    {
      cout << "net counts : " << endl;
      cout << "\t\ttrain count\t\t\t\tinferred count" << endl << endl;
    }

    for (int clauseno = 0; clauseno < clauseCnt; clauseno++) 
    {
      if (!relevantClauses[clauseno]) continue;
      
      if (vpdebug)
        cout << clauseno << ":\t\t" << clauseTrainCnts[clauseno] << "\t\t\t\t"
             << clauseInferredCnts[clauseno] << endl;
      if (rescaleGradient_ && clauseTrainCnts[clauseno] > 0)
      {
        gradient[clauseno] += 
          (clauseTrainCnts[clauseno] - clauseInferredCnts[clauseno])
            / clauseTrainCnts[clauseno];
      }
      else
      {
        gradient[clauseno] += clauseTrainCnts[clauseno] - 
                              clauseInferredCnts[clauseno];
      }
    }

    delete[] clauseTrainCnts;
    delete[] clauseInferredCnts;
  }


    // Get the gradient 
  void getGradient(double* const & weights, double* const & gradient,
                   const int numWts) 
  {
    // Set the weights and run inference
    
    //cout << "New Weights = **** " << endl << endl;
    
      // If there is one db or the clauses for multiple databases line up
    if (idxTrans_ == NULL)
    {
      int clauseCnt = clauseCntPerDomain_[0];
      for (int i = 0; i < domains_.size(); i++)
      {
        Array<bool>& relevantClauses = relevantClausesPerDomain_[i];
        assert(clauseCntPerDomain_[i] == clauseCnt);
        MLN* mln = mlns_[i];
        
        for (int j = 0; j < clauseCnt; j++) 
        {
          Clause* c = (Clause*) mln->getClause(j);
          if (relevantClauses[j]) c->setWt(weights[j]);
          else                    c->setWt(0);
        }
      }
    }
    else
    {   // The clauses for multiple databases do not line up
      Array<Array<double> >* wtsPerDomain = idxTrans_->getWtsPerDomain();
      const Array<Array<Array<IdxDiv>*> >* cIdxToCFIdxsPerDomain 
        = idxTrans_->getClauseIdxToClauseFormulaIdxsPerDomain();
      
      for (int i = 0; i < domains_.size(); i++)
      {
        Array<double>& wts = (*wtsPerDomain)[i];
        memset((double*)wts.getItems(), 0, wts.size()*sizeof(double));

          //map clause/formula weights to clause weights
        for (int j = 0; j < wts.size(); j++)
        {
          Array<IdxDiv>* idxDivs = (*cIdxToCFIdxsPerDomain)[i][j];          
          for (int k = 0; k < idxDivs->size(); k++)
            wts[j] += weights[ (*idxDivs)[k].idx ] / (*idxDivs)[k].div;
        }
      }
      
      for (int i = 0; i < domains_.size(); i++)
      {
        Array<bool>& relevantClauses = relevantClausesPerDomain_[i];
        int clauseCnt = clauseCntPerDomain_[i];
        Array<double>& wts = (*wtsPerDomain)[i];
        assert(wts.size() == clauseCnt);
        MLN* mln = mlns_[i];

        for (int j = 0; j < clauseCnt; j++)
        {
          Clause* c = (Clause*) mln->getClause(j);
          if(relevantClauses[j]) c->setWt(wts[j]);
          else                   c->setWt(0);
        }
      }
    }
    //for (int i = 0; i < numWts; i++) cout << i << " : " << weights[i] << endl;

    if (withEM_) fillInMissingValues();
    cout << "Running inference ..." << endl;
    infer();
    cout << "Done with inference" << endl;

      // Compute the gradient
    memset(gradient, 0, numWts*sizeof(double));

      // There is one DB or the clauses of multiple DBs line up
    if (idxTrans_ == NULL)
    {
      for(int i = 0; i < domainCnt_; i++) 
      {		  
        //cout << "For domain number " << i << endl << endl; 
        getGradientForDomain(gradient, i);        
      }
    }
    else
    {
        // The clauses for multiple databases do not line up
      Array<Array<double> >* gradsPerDomain = idxTrans_->getGradsPerDomain();
      const Array<Array<Array<IdxDiv>*> >* cIdxToCFIdxsPerDomain 
        = idxTrans_->getClauseIdxToClauseFormulaIdxsPerDomain();
     
      for (int i = 0; i < domainCnt_; i++) 
      {		  
        //cout << "For domain number " << i << endl << endl; 

        Array<double>& grads = (*gradsPerDomain)[i];
        memset((double*)grads.getItems(), 0, grads.size()*sizeof(double));
        
        getGradientForDomain((double*)grads.getItems(), i);
        
          // map clause gradient to clause/formula gradients
        assert(grads.size() == clauseCntPerDomain_[i]);
        for (int j = 0; j < grads.size(); j++)
        {
          Array<IdxDiv>* idxDivs = (*cIdxToCFIdxsPerDomain)[i][j];          
          for (int k = 0; k < idxDivs->size(); k++)
            gradient[ (*idxDivs)[k].idx ] += grads[j] / (*idxDivs)[k].div;
        }
      }
    }

      // Add the deriative of the prior 
    if (usePrior_) 
    {
	  for (int i = 0; i < numWts; i++) 
      {
        if (!relevantClausesFormulas_[i]) continue;
        double priorDerivative = -(weights[i]-priorMeans_[i])/
                                 (priorStdDevs_[i]*priorStdDevs_[i]);
        //cout << i << " : " << "gradient : " << gradient[i]
        //     << "  prior gradient : " << priorDerivative;
        gradient[i] += priorDerivative; 
	//cout << "  net gradient : " << gradient[i] << endl; 
      }
    }
  }


 private:
  int domainCnt_;
  Array<Domain*> domains_;  
  Array<MLN*> mlns_;
  Array<Array<double> > logOddsPerDomain_;
  Array<int> clauseCntPerDomain_;

	// Used in lazy version
  Array<Array<double> > totalTrueCnts_; 
  Array<Array<double> > defaultTrueCnts_;

  Array<Array<bool> > relevantClausesPerDomain_;
  Array<bool> relevantClausesFormulas_;

    // Predicates corresponding to the groundings of the non-evidence
    // predicates
  //Array<GroundPredicateHashArray> gndPredsPerDomain_;
   
    // Actual truth values of the non-evidence ground preds
  //Array<Array<TruthValue> > gndPredValuesPerDomain_;
   
	// Used to compute cnts from mrf
  Array<double*> trainTrueCnts_;

  bool usePrior_;
  const double* priorMeans_, * priorStdDevs_; 

  MaxWalksatParams* wsparams_;
  IndexTranslator* idxTrans_; //not owned by object; don't delete
  
  int memLimit_;
  bool lazyInference_;
  bool lazyReset_;
  bool rescaleGradient_;

    // Replace MRFs with these
  Array<VariableState*> states_;
  Array<Inference*> inferences_;
  
    // Using EM to fill in missing values?
  bool withEM_;
    // Used to fill in missing values
  Array<VariableState*> mvStates_;
  Array<Inference*> mvInferences_;

    // Predicates corresponding to the groundings of the unknown evidence
    // predicates
  Array<GroundPredicateHashArray> uePredsPerDomain_;


   // Predicates corresponding to the groundings of the known evidence
   // predicates
  Array<GroundPredicateHashArray> kePredsPerDomain_;
    // Actual truth values of ground known evidence preds
  Array<Array<TruthValue> > kePredValuesPerDomain_;


    // Predicates corresponding to the groundings of the unknown non-evidence
    // predicates
  Array<GroundPredicateHashArray> unePredsPerDomain_;
    // Predicates corresponding to the groundings of the unknown non-evidence
    // predicates copy of unePredsPerDomain_ for the MRFs to infer missing
    // values
  Array<GroundPredicateHashArray> uneEMPredsPerDomain_;


    // Predicates corresponding to the groundings of the known non-evidence
    // predicates
  Array<GroundPredicateHashArray> knePredsPerDomain_;
    // Actual truth values of ground known non-evidence preds
  Array<Array<TruthValue> > knePredValuesPerDomain_;  

};


#endif
